/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.reports.availability;

import com.net2plan.interfaces.networkDesign.IReport;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.interfaces.simulation.IEventProcessor;
import com.net2plan.interfaces.simulation.SimEvent;
import com.net2plan.libraries.SRGUtils;
import com.net2plan.utils.*;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.*;
import javax.xml.stream.XMLOutputFactory;
import org.codehaus.stax2.XMLOutputFactory2;
import org.codehaus.stax2.XMLStreamWriter2;

/**
 * This report analyzes a network design in terms of average availability under a set 
 * of pre-defined failures in the form of Shared-Risk Groups (SRGs).
 * 
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class Report_availability implements IReport
{
	private Map<Long, Map<Long, Double>> A_d_0_1, A_d_0_2, A_d_1, A_d_2;
	private Map<Long, Double> A_D_0_1, A_D_0_2, A_D_1, A_D_2;
	private double pi_excess;
	
	@Override
	public String executeReport(NetPlan netPlan, Map<String, String> reportParameters, Map<String, String> net2planParameters)
	{
		boolean analyzeDoubleFailures, assumeUnprotectedRoutes, omitProtectionSegments;
		double defaultMTTFInHours;
		double defaultMTTRInHours;
		String failureModel;

		/* Check report parameters */
		try { analyzeDoubleFailures = Boolean.parseBoolean(reportParameters.get("analyzeDoubleFailures")); }
		catch (Exception ex) { throw new Net2PlanException("Parameter 'analyzeDoubleFailures' value must be true or false"); }

		try { assumeUnprotectedRoutes = Boolean.parseBoolean(reportParameters.get("assumeUnprotectedRoutes")); }
		catch (Exception ex) { throw new Net2PlanException("Parameter 'assumeUnprotectedRoutes' value must be true or false"); }

		try { omitProtectionSegments = Boolean.parseBoolean(reportParameters.get("omitProtectionSegments")); }
		catch (Exception ex) { throw new Net2PlanException("Parameter 'omitProtectionSegments' value must be true or false"); }

		try
		{
			defaultMTTFInHours = Double.parseDouble(reportParameters.get("defaultMTTFInHours"));
			if (defaultMTTFInHours <= 0) defaultMTTFInHours = Double.MAX_VALUE;
		}
		catch (Exception ex)
		{
			throw new Net2PlanException("Parameter 'defaultMTTFInHours' value must be a valid number");
		}

		try
		{
			defaultMTTRInHours = Double.parseDouble(reportParameters.get("defaultMTTRInHours"));
			if (defaultMTTRInHours <= 0) throw new Exception("Bad");
		}
		catch (Exception ex)
		{
			throw new Net2PlanException("Parameter 'defaultMTTRInHours' value must be greater than zero");
		}

		failureModel = reportParameters.get("failureModel");

		String algorithmFile = reportParameters.get("provisioningAlgorithm_file");
		String algorithmName = reportParameters.get("provisioningAlgorithm_classname");
		String algorithmParam = reportParameters.get("provisioningAlgorithm_parameters");
		if (algorithmFile.isEmpty() || algorithmName.isEmpty()) throw new Net2PlanException("A provisioning algorithm must be defined");

		Map<String, String> algorithmParameters = StringUtils.stringToMap(algorithmParam);

		switch (failureModel)
		{
			case "SRGfromNetPlan":
				break;

			case "perNode":
				SRGUtils.configureSRGs(netPlan, defaultMTTFInHours, defaultMTTRInHours, SRGUtils.SharedRiskModel.PER_NODE, true);
				break;

			case "perLink":
				SRGUtils.configureSRGs(netPlan, defaultMTTFInHours, defaultMTTRInHours, SRGUtils.SharedRiskModel.PER_LINK, true);
				break;

			case "perDirectionalLinkBundle":
				SRGUtils.configureSRGs(netPlan, defaultMTTFInHours, defaultMTTRInHours, SRGUtils.SharedRiskModel.PER_DIRECTIONAL_LINK_BUNDLE, true);
				break;

			case "perBidirectionalLinkBundle":
				SRGUtils.configureSRGs(netPlan, defaultMTTFInHours, defaultMTTRInHours, SRGUtils.SharedRiskModel.PER_BIDIRECTIONAL_LINK_BUNDLE, true);
				break;

			default:
				throw new Net2PlanException("Failure model not valid. Please, check algorithm parameters description");
		}

		Map<Long, Double> A_f = netPlan.getSRGAvailabilityMap();
		Set<Long> srgIds = netPlan.getSRGIds();
		List<Set<Long>> F_s = defineFailureStates(srgIds, analyzeDoubleFailures);
		
		/* Compute state probabilities (pi_s) */
		List<Double> pi_s = defineStateProbabilities(F_s, A_f);
		double sum_pi_s = DoubleUtils.sum(pi_s);
		pi_excess = 1 - sum_pi_s;
		
		A_d_0_1 = new LinkedHashMap<Long, Map<Long, Double>>();
		A_d_0_2 = new LinkedHashMap<Long, Map<Long, Double>>();
		A_d_1 = new LinkedHashMap<Long, Map<Long, Double>>();
		A_d_2 = new LinkedHashMap<Long, Map<Long, Double>>();
		A_D_0_1 = new LinkedHashMap<Long, Double>();
		A_D_0_2 = new LinkedHashMap<Long, Double>();
		A_D_1 = new LinkedHashMap<Long, Double>();
		A_D_2 = new LinkedHashMap<Long, Double>();
		Set<Long> layerIds = netPlan.getLayerIds();
		for(long layerId : layerIds)
		{
			if (omitProtectionSegments) netPlan.removeAllProtectionSegments(layerId);
			
			A_d_0_1.put(layerId, new LinkedHashMap<Long, Double>());
			A_d_0_2.put(layerId, new LinkedHashMap<Long, Double>());
			A_d_1.put(layerId, new LinkedHashMap<Long, Double>());
			A_d_2.put(layerId, new LinkedHashMap<Long, Double>());
			
			double H_d = 0;
			double BlockedTraffic_d = 0;
			Set<Long> demandIds = netPlan.getDemandIds(layerId);
			for(long demandId : demandIds)
			{
				double h_d = netPlan.getDemandOfferedTraffic(layerId, demandId);
				double blockedTraffic_d = netPlan.getDemandBlockedTraffic(layerId, demandId);
				
				A_d_0_1.get(layerId).put(demandId, blockedTraffic_d > 0 ? 0.0 : 1.0);
				A_d_0_2.get(layerId).put(demandId, h_d == 0 ? 0 : Math.max(0, 1 - blockedTraffic_d / h_d));
				A_d_1.get(layerId).put(demandId, 0.0);
				A_d_2.get(layerId).put(demandId, 0.0);
				
				H_d += h_d;
				BlockedTraffic_d += blockedTraffic_d;
			}
			
			A_D_0_1.put(layerId, BlockedTraffic_d > 0 ? 0.0 : 1.0);
			A_D_0_2.put(layerId, H_d == 0 ? 0 : Math.max(0, 1 - BlockedTraffic_d / H_d));
			A_D_1.put(layerId, 0.0);
			A_D_2.put(layerId, 0.0);
		}

		Iterator<Set<Long>> stateIterator = F_s.iterator();
		Iterator<Double> pi_s_iterator = pi_s.iterator();
		while (stateIterator.hasNext())
		{
			Set<Long> srgIds_thisState = stateIterator.next();
			double pi_s_thisState = pi_s_iterator.next();

			NetPlan currentNetPlan;
			if (srgIds_thisState.isEmpty())
			{
				currentNetPlan = netPlan;
			}
			else
			{
				currentNetPlan = netPlan.copy();

				if (assumeUnprotectedRoutes)
				{
					/* Compute the effects of the failure */
					for (long srgId : srgIds_thisState)
						currentNetPlan.setSRGDown(srgId);
				}
				else
				{
					/* Initialize the reaction algorithm */
					IEventProcessor algorithm = ClassLoaderUtils.getInstance(new File(algorithmFile), algorithmName, IEventProcessor.class);
					algorithm.initialize(currentNetPlan, currentNetPlan, algorithmParameters, reportParameters, net2planParameters);
					
					/* Compute the effects of the failure */
					currentNetPlan.activateActionMonitoring();
					for (long srgId : srgIds_thisState)
						currentNetPlan.setSRGDown(srgId);
					SimEvent event = new SimEvent(0, currentNetPlan.getActionsPerformed(), SimEvent.DestinationModule.EVENT_PROCESSOR);

					try
					{
						/* Apply the reaction algorithm */
						algorithm.processEvent(currentNetPlan, event);
					}
					catch (Throwable e)
					{
						try { ((Closeable) algorithm.getClass().getClassLoader()).close(); }
						catch (Throwable e1) { }					

						throw (e);
					}
					
					try { ((Closeable) algorithm.getClass().getClassLoader()).close(); }
					catch (Throwable e1) { }					
				}
			}

			for(long layerId : layerIds)
			{
				double H_d = 0;
				double BlockedTraffic_d = 0;
				Set<Long> demandIds = netPlan.getDemandIds(layerId);
				for(long demandId : demandIds)
				{
					double h_d = netPlan.getDemandOfferedTraffic(layerId, demandId);
					H_d += h_d;

					if (!currentNetPlan.isDemandActive(layerId, demandId)) continue;
					double blockedTraffic_d = currentNetPlan.getDemandBlockedTraffic(layerId, demandId);
					BlockedTraffic_d += blockedTraffic_d;

					A_d_1.get(layerId).put(demandId, A_d_1.get(layerId).get(demandId) + (blockedTraffic_d > 0 ? 0.0 : pi_s_thisState));
					A_d_2.get(layerId).put(demandId, A_d_2.get(layerId).get(demandId) + (h_d == 0 ? 0 : pi_s_thisState * Math.max(0, 1 - blockedTraffic_d / h_d)));
				}

				A_D_1.put(layerId, A_D_1.get(layerId) + (BlockedTraffic_d > 0 ? 0.0 : pi_s_thisState));
				A_D_2.put(layerId, A_D_2.get(layerId) + (H_d == 0 ? 0 : pi_s_thisState * Math.max(0, 1 - BlockedTraffic_d / H_d)));
			}
		}

		return generateReport(netPlan);
	}

	@Override
	public String getDescription()
	{
		return "This report analyzes a network design in terms of average availability under a set of pre-defined failures in the form of Shared-Risk Groups (SRGs)";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> reportParameters = new LinkedList<Triple<String, String, String>>();
		reportParameters.add(Triple.of("provisioningAlgorithm", "#eventProcessor#", "Algorithm to process failure events"));
		reportParameters.add(Triple.of("analyzeDoubleFailures", "#boolean# true", "Indicates whether double failures are studied"));
		reportParameters.add(Triple.of("assumeUnprotectedRoutes", "#boolean# false", "Do not apply the recovery algorithm"));
		reportParameters.add(Triple.of("defaultMTTFInHours", "8748", "Default value for Mean Time To Fail (hours)"));
		reportParameters.add(Triple.of("defaultMTTRInHours", "12", "Default value for Mean Time To Repair (hours)"));
		reportParameters.add(Triple.of("failureModel", "#select# perBidirectionalLinkBundle SRGfromNetPlan perNode perLink perDirectionalLinkBundle", "Failure model selection: SRGfromNetPlan, perNode, perLink, perDirectionalLinkBundle, perBidirectionalLinkBundle"));
		reportParameters.add(Triple.of("omitProtectionSegments", "#boolean# false", "Remove protection segments from the network plan to free their reserved bandwidth"));

		return reportParameters;
	}

	@Override
	public String getTitle()
	{
		return "Availability report";
	}

	/**
	 * Returns the set of SRGs going down on each failure state.
	 * 
	 * @param srgIds Set of SRG identifiers
	 * @param analyzeDoubleFailures Flag to indicate whether or not double failure states are included
	 * @return Set of SRGs going down on each failure state
	 * @since 1.1
	 */
	protected List<Set<Long>> defineFailureStates(Set<Long> srgIds, boolean analyzeDoubleFailures)
	{
		return SRGUtils.enumerateFailureStates(srgIds, true, analyzeDoubleFailures);
	}

	/**
	 * Computes the probability to find the network on each failure state.
	 * 
	 * @param F_s Set of SRGs failing per each failure state
	 * @param A_f Availability value per SRG
	 * @return Probability to find the network in each failure state
	 * @since 1.1
	 */
	protected List<Double> defineStateProbabilities(List<Set<Long>> F_s, Map<Long, Double> A_f)
	{
		return SRGUtils.computeStateProbabilities(F_s, A_f);
	}

	/**
	 * Returns an HTML {@code String} containing the report.
	 * 
	 * @param netPlan A network plan
	 * @return Report in HTML format
	 * @since 1.1
	 */
	private String generateReport(NetPlan netPlan)
	{
		try (ByteArrayOutputStream os = new ByteArrayOutputStream())
		{
			XMLOutputFactory2 output = (XMLOutputFactory2) XMLOutputFactory.newFactory();
			XMLStreamWriter2 writer = (XMLStreamWriter2) output.createXMLStreamWriter(os);
			
			writer.writeStartDocument("UTF-8", "1.0");
			writer.writeStartElement("network");
			
			/* Write layer information */
			Set<Long> layerIds = netPlan.getLayerIds();
			for(long layerId : layerIds)
			{
				writer.writeStartElement("layer");
				writer.writeAttribute("id", Long.toString(layerId));
				writer.writeAttribute("name", netPlan.getLayerName(layerId));
				writer.writeAttribute("availabilityClassicPesimistic", String.format("%.6f", A_D_1.get(layerId)));
				writer.writeAttribute("availabilityClassicOptimistic", String.format("%.6f", A_D_1.get(layerId) + pi_excess * A_D_0_1.get(layerId)));
				writer.writeAttribute("availabilityWeightedPesimistic", String.format("%.6f", A_D_2.get(layerId)));
				writer.writeAttribute("availabilityWeightedOptimistic", String.format("%.6f", A_D_2.get(layerId) + pi_excess * A_D_0_2.get(layerId)));
				writer.writeAttribute("totalOfferedTraffic", String.format("%.3f", netPlan.getDemandTotalOfferedTraffic(layerId)));
				writer.writeAttribute("totalFailureFreeBlockedTraffic", String.format("%.3f", netPlan.getDemandTotalBlockedTraffic(layerId)));

				/* Write demand information */
				Set<Long> demandIds = netPlan.getDemandIds(layerId);
				if (demandIds.isEmpty())
				{
					writer.writeAttribute("worstDemandAvailabilityClassicPesimistic", Double.toString(0.0));
					writer.writeAttribute("worstDemandAvailabilityClassicOptimistic", Double.toString(0.0));
					writer.writeAttribute("worstDemandAvailabilityWeightedPesimistic", Double.toString(0.0));
					writer.writeAttribute("worstDemandAvailabilityWeightedOptimistic", Double.toString(0.0));
				}
				else
				{
					long worstD_1 = DoubleUtils.minIndexes(A_d_1.get(layerId), Constants.SearchType.FIRST).iterator().next();
					long worstD_2 = DoubleUtils.minIndexes(A_d_2.get(layerId), Constants.SearchType.FIRST).iterator().next();
					writer.writeAttribute("worstDemandAvailabilityClassicPesimistic", String.format("%.6f", A_d_1.get(layerId).get(worstD_1)));
					writer.writeAttribute("worstDemandAvailabilityClassicOptimistic", String.format("%.6f", A_d_1.get(layerId).get(worstD_1) + pi_excess * A_d_0_1.get(layerId).get(worstD_1)));
					writer.writeAttribute("worstDemandAvailabilityWeightedPesimistic", String.format("%.6f", A_d_2.get(layerId).get(worstD_2)));
					writer.writeAttribute("worstDemandAvailabilityWeightedOptimistic", String.format("%.6f", A_d_2.get(layerId).get(worstD_2) + pi_excess * A_d_0_2.get(layerId).get(worstD_2)));

					for(long demandId : demandIds)
					{
						long ingressNodeId_thisDemand = netPlan.getDemandIngressNode(layerId, demandId);
						long egressNodeId_thisDemand = netPlan.getDemandEgressNode(layerId, demandId);
						String ingressNodeName = netPlan.getNodeName(ingressNodeId_thisDemand);
						String egressNodeName = netPlan.getNodeName(egressNodeId_thisDemand);
						double h_d = netPlan.getDemandOfferedTraffic(layerId, demandId);
						double blockedTraffic_d = netPlan.getDemandBlockedTraffic(layerId, demandId);

						writer.writeStartElement("demand");
						writer.writeAttribute("id", Long.toString(demandId));
						writer.writeAttribute("ingressNode", ingressNodeName.isEmpty() ? Long.toString(ingressNodeId_thisDemand) : String.format("%d (%s)", ingressNodeId_thisDemand, ingressNodeName));
						writer.writeAttribute("egressNode", egressNodeName.isEmpty() ? Long.toString(egressNodeId_thisDemand) : String.format("%d (%s)", egressNodeId_thisDemand, egressNodeName));
						writer.writeAttribute("offeredTraffic", String.format("%.3f", h_d));
						writer.writeAttribute("failureFreeBlockedTraffic", String.format("%.3f", blockedTraffic_d));
						writer.writeAttribute("availabilityClassicPesimistic", String.format("%.6f", A_d_1.get(layerId).get(demandId)));
						writer.writeAttribute("availabilityClassicOptimistic", String.format("%.6f", A_d_1.get(layerId).get(demandId) + pi_excess * A_d_0_1.get(layerId).get(demandId)));
						writer.writeAttribute("availabilityWeightedPesimistic", String.format("%.6f", A_d_2.get(layerId).get(demandId)));
						writer.writeAttribute("availabilityWeightedOptimistic", String.format("%.6f", A_d_2.get(layerId).get(demandId) + pi_excess * A_d_0_2.get(layerId).get(demandId)));
						writer.writeEndElement();
					}
				}
				
				writer.writeEndElement();
			}				
				
			writer.writeEndElement();
			writer.writeEndDocument();
			writer.flush();
			writer.close();

			String xml = os.toString(StandardCharsets.UTF_8.name());
			return HTMLUtils.getHTMLFromXML(xml, getReportTemplateFile());
		}
		catch(Throwable e)
		{
			throw new RuntimeException(e);
		}
	}

	/**
	 * Returns the location of the template file for the report.
	 * 
	 * @return Location of the template file
	 * @since 1.1
	 */
	protected URL getReportTemplateFile()
	{
		try { return Report_availability.class.getResource("/com/net2plan/examples/reports/availability/Report_availability.xsl").toURI().toURL(); }
		catch (URISyntaxException | MalformedURLException ex) { throw new RuntimeException(ex); }
	}
}